import unittest
import numpy as np
from worker.data import math
from worker.data import operations


class TestGeneral(unittest.TestCase):
    def test_operations(self):
        self.assertFalse(operations.is_number("nan"))
        self.assertFalse(operations.is_number(None))
        self.assertFalse(operations.is_number(""))
        self.assertFalse(operations.is_number(''))
        self.assertFalse(operations.is_number("null"))
        self.assertFalse(operations.is_number(np.nanpercentile([np.nan], 50)))
        self.assertTrue(operations.is_number(-999.25))
        self.assertTrue(operations.is_number("-999.25"))
        self.assertTrue(operations.is_number("0.0"))

        self.assertTrue(operations.is_number(float('inf')))
        self.assertFalse(operations.is_finite(float('inf')))

        self.assertTrue(operations.is_finite(1.1))
        self.assertFalse(operations.is_finite("1.1"))

        self.assertIsNone(operations.to_number("nan"))
        self.assertIsNone(operations.to_number(""))
        self.assertIsNone(operations.to_number(''))
        self.assertIsNone(operations.to_number("null"))
        self.assertEqual(operations.to_number(-999.25), -999.25)
        self.assertEqual(operations.to_number("-999.25"), -999.25)
        self.assertEqual(operations.to_number("0.0"), 0.0)

        np.testing.assert_equal(operations.none_to_nan(None), np.nan)
        np.testing.assert_equal(operations.none_to_nan(np.nan), np.nan)
        np.testing.assert_equal(operations.none_to_nan(1), 1)
        np.testing.assert_equal(operations.none_to_nan([1, 2, None]), [1, 2, np.nan])
        np.testing.assert_equal(operations.none_to_nan([1, 2, 3, np.nan, None]), [1, 2, 3, np.nan, np.nan])

    def test_fetch_from_path(self):
        self.assertTrue(operations.get_data_by_path(dict(name="X", value=2.3), 'value', float), 2.3)
        self.assertTrue(operations.get_data_by_path(dict(name="X", value=2.3), 'property', float, default=1.0), 1.0)

        self.assertTrue(operations.get_data_by_path(dict(name={'first_name': 'foo', 'last_name': 'bar'}, value=2.3), 'name.first_name'), 'foo')
        self.assertTrue(operations.get_data_by_path(dict(p1={'p2': 1.2}), 'p1.p2'), 1.2)

        # if default is not provided and the key can not be found it will raise an exception
        self.assertRaises(Exception, operations.get_data_by_path, {"key": "value"}, "a_path")
        # if default is provided (even None) and the key can not be found it return the default value
        self.assertIsNone(operations.get_data_by_path({"key": "value"}, "a_path", default=None))
        self.assertEqual("something", operations.get_data_by_path(data={"key": "value"}, path="a_path", default="something"))

    def test_math_utils(self):
        self.assertEqual(math.percentile([1, 2, None, np.nan], 50), 1.5)
        self.assertEqual(math.percentile([None, 1, np.nan], 50), 1.0)
        self.assertIsNone(math.percentile([np.nan, np.nan], 50))
        self.assertIsNone(math.percentile([None], 50))

        mean_angle = math.mean_angles([248, 315, 174, 112, 236, 276, 276, 39, 270, 231, 259, 186, 298])
        self.assertTrue(253 < mean_angle < 254)
        self.assertTrue(2.999 <= math.mean_angles([1, 2, 3, 4, 5]) <= 3.001)
        self.assertTrue(356.999 <= math.mean_angles([-1, -2, -3, -4, -5]) <= 357.001)

        self.assertEqual(math.angle_difference(10, 350), -20.0)
        self.assertEqual(math.angle_difference(350, 10), 20.0)

        self.assertEqual(math.abs_angle_difference(10, 350), 20.0)
        self.assertEqual(math.abs_angle_difference(350, 10), 20.0)

        self.assertEqual(
            math.split_zip_edges(np.array([0, 1, 2, 4, 6, 8, 9, 10, 12, 14, 15, 16])),
            [(0, 2), (4, 4), (6, 6), (8, 10), (12, 12), (14, 16)]
        )
        self.assertEqual(
            math.split_zip_edges(np.array([0, 1, 2, 4, 6, 8, 9, 10, 12, 14, 15, 16]), min_segment_length=2),
            [(0, 2), (8, 10), (14, 16)]
        )

        self.assertEqual(
            math.start_stop([1, 1, 1, 2, 4, 1, 1, 1, 2, 1, 1, 1], 1),
            [(0, 2), (5, 7), (9, 11)]
        )
        self.assertEqual(
            math.start_stop([True, True, True, False, False, True, True, True, False, True, True, True], True),
            [(0, 2), (5, 7), (9, 11)]
        )



import functools
from wrapt_timeout_decorator import timeout


def timeout_track(before=2, default_timeout=30):
    """
    A decorator that can be applied to a lambda function handler
    to raise an exception just before a timeout occur
    :param before: the seconds before the timeout occurs; used with lambda function
    :param default_timeout: actual timeout in seconds with cases other than lambda function
    :return:
    """
    def wrapper(func):
        @functools.wraps(func)
        @timeout(
            dec_timeout=f"((args[1].get_remaining_time_in_millis() - 1000 * {before}) / 1000) if args[1] else {default_timeout}",
            dec_allow_eval=True
        )
        def decorated(*args, **kwargs):
            return func(*args, **kwargs)

        return decorated
    return wrapper



from testcase.app.drilling_efficiency_app import DrillingEfficiency


def lambda_handler(event, context):
    """
    This function is the main entry point of the AWS Lambda function
    :param event: a scheduler or kafka event
    :param context: AWS Context
    :return:
    """
    app = DrillingEfficiency()
    app.load(event)
    app.run_modules()
    app.save_state()



import json
import os

from worker import API
from worker.mixins.logging import LoggingMixin
from worker.mixins.rollbar import RollbarMixin
from worker.data.enums import LambdaStates


class TaskHandler(LoggingMixin, RollbarMixin):
    def __init__(self, *args, **kwargs) -> None:
        self._api = kwargs.pop("api", API())
        super().__init__(*args, **kwargs)
        self.event = {}
        self.task_id = ""
        self.task = {}

    def process(self, *args, **kwargs):
        """
        Provides a placeholder for actual logic code for each app.
        e.g. Instantiate the app and run it
        This method must be overriden by each app
        :return:
        """
        raise NotImplementedError("process method must be defined before instantiating this class")

    def run_process(self, event: dict, *args, **kwargs):
        """
        Runs the task handler workflow.
        This method would be called after instantiation of the TaskHandler class
        :param event:
        :return:
        """
        self.event = event
        self.task_id = self.event.get("task_id")
        self.get_task(self.task_id)

        # AWS_EXECUTION_ENV is a reserved environment variable it will hold the runtime environment eg python3.6, java11 etc.
        # We do not have to set it on our lambda functions, it will be set by AWS
        # We can use this to check if we are running on hosted or local
        is_hosted = os.environ.get("AWS_EXECUTION_ENV", False)

        test_mode = self.task.get("properties", {}).get("test_mode", False)

        if test_mode and is_hosted:
            asset_id = self.task.get("asset_id", 0)
            self.warn(asset_id, f"Not processing on AWS due to test mode. task_id: {self.task_id} maybe running.")
            return

        try:
            data = self.process(self.event, *args, **kwargs)
            request_body = {
                "task": {
                    "payload": {
                        "data": data,
                        "message": LambdaStates.SUCCEEDED.value
                    }
                }
            }
            self.update_task(self.task_id, "success", request_body)
            return data

        except Exception as exception:
            request_body = {
                "task": {
                    "fail_reason": f"Message: {LambdaStates.FAILED.value}, Exception: {exception}"
                }
            }
            self.update_task(self.task_id, "fail", request_body)
            raise exception

    def get_task(self, task_id: str) -> None:
        """
        Gets the task from API endpoint for the given task_id
        :param task_id:
        :return:
        """
        path = "/v2/tasks/{0}".format(task_id)
        self.task = self._api.get(path).data

    def update_task(self, task_id: str, status: str, request_body: dict) -> None:
        """
        Updates the task status on the API endpoint in case of success or failure
        :param task_id:
        :param status: [success, fail]
        :param request_body:
        :return:
        """
        path = "/v2/tasks/{0}/{1}".format(task_id, status)
        self.task = self._api.put(path, data=json.dumps(request_body)).data



import importlib
import worker
importlib.reload(worker)
from worker import constants  # noqa: #402


constants.update({
    "global": {
        "app-name": "WorkerTest-DrillingEfficiency",
        "app-key": "drilling-efficiency",
        "event-type": "scheduler",
        "query-limit": 3600,
    },
    "drilling-efficiency": {
        "mse": {
            "export-duration": 30,  # update and export results every 30 seconds
            "drilling-activity": ["Rotary Drilling", "Slide Drilling"],
            "required-channels": ['weight_on_bit', 'rotary_rpm', 'rotary_torque', 'rop'],
            "running-string": ["drillstring"],
            "reset-config": ["drillstring"]
        }
    }
})



from testcase.state_testing_app.state_modules import TestRedis, TestMongo, TestMigration
from worker import App


class TestingApp(App):
    # override
    def get_modules(self):
        return [
            TestMongo,  # Mongo testing module
            TestRedis,  # Redis testing module
            TestMigration  # Migration from one state storage to another testing module
        ]

    def load(self, event):
        return



import numpy as np
from worker.data.operations import none_to_nan, is_number


def percentile(ls, percent):
    """
    Calculates percentile without considering np.nan and None values. Returns 0 is all values are np.nan or None
    :param ls:
    :param percent:
    :return:
    """
    try:
        p = np.nanpercentile(none_to_nan(ls), percent)
        if p >= 0 or p <= 0:
            return p

        return None
    except TypeError:
        return None


def mean_angles(ls):
    """
    to compute mean of list of angles
    :param ls: a list of angles
    :return: mean value of angles
    """
    x_mean = np.nanmean(np.cos(np.deg2rad(ls)))
    y_mean = np.nanmean(np.sin(np.deg2rad(ls)))
    mean_deg = np.rad2deg(np.arctan2(y_mean, x_mean))

    return mean_deg % 360


def angle_difference(ang1, ang2):
    """
    Code from: https://rosettacode.org/wiki/Angle_difference_between_two_bearings#Python
    :param ang1:
    :param ang2:
    :return:
    """
    if not is_number(ang1) or not is_number(ang2):
        return None

    r = (ang2 - ang1) % 360.0
    if r >= 180.0:
        r -= 360.0
    return r


def abs_angle_difference(ang1, ang2):
    """
    The absolute difference between two angles.
    :param ang1:
    :param ang2:
    :return:
    """
    diff = angle_difference(ang1, ang2)
    if is_number(diff):
        return abs(diff)

    return None


def split_zip_edges(arr, separation_length=1, min_segment_length=1):
    """
    In cases that you have elements and you want only values that are close to each other.
    :param arr: array of non-continuous data
    :param separation_length: separation length
    :param min_segment_length: min length of each segment
    :return: a list of tuples representing the start and stop of each segment
    """
    if isinstance(arr, list):
        arr = np.array(arr)

    m = np.concatenate(([True], arr[1:] > arr[:-1] + separation_length, [True]))
    idx = np.flatnonzero(m)
    ll = arr.tolist()
    return [(ll[i], ll[j - 1]) for i, j in zip(idx[:-1], idx[1:]) if (ll[j - 1] + 1 - ll[i]) >= min_segment_length]


def start_stop(arr, trigger_val, min_len_thresh=1):
    """
    If you have an array representing values and you only want
    the values which are equal to a specific trigger_value.
    Another param is the minimum of each interval size.
    :param arr: a continuous stream of data
    :param trigger_val: desired value
    :param min_len_thresh: the min distance between two separate segments
    :return: a list of tuples representing the start and stop of each segment
    """
    # "Enclose" mask with sentient to catch shifts later on
    mask = np.r_[False, np.equal(arr, trigger_val), False]

    # Get the shifting indices
    idx = np.flatnonzero(mask[1:] != mask[:-1])

    # Get lengths
    lens = idx[1::2] - idx[::2]

    res = idx.reshape(-1, 2)[lens >= min_len_thresh] - [0, 1]

    return [(i[0], i[-1]) for i in res]
